/*
 * pcap.xs
 *
 * XS wrapper for LBL pcap(3) library.
 *
 * Copyright (c) 1999 Tim Potter. All rights reserved. This program is free
 * software; you can redistribute it and/or modify it under the same terms
 * as Perl itself.
 *
 * $Id: Pcap.xs,v 1.11 2000/05/15 05:02:50 tpot Exp $
 *
 * Patched for Win32 - JLM - 2004/09/03
 */

#ifdef __cplusplus
extern "C" {
#endif

#include "EXTERN.h"
#define  WINSOCK2_H_REQUESTED
#include "perl.h"
#include "XSUB.h"

#include <pcap.h>

#ifdef _WIN32
#define _WINPCAP
#include <Win32-Extensions.h>
#endif

#ifdef _WINPCAP
#include <remote-ext.h>
#endif

/* avoid space in the name of struct_bpf_programPtr class */
typedef struct bpf_program struct_bpf_program;

#ifdef __cplusplus
}
#endif

/* The following taken from ext/Data/Dumper/Dumper.c in the Perl 5.6.0
   distribution. */

#if PERL_VERSION < 5
#  ifndef PL_sv_undef
#    define PL_sv_undef sv_undef
#  endif
#  ifndef PL_na
#    define PL_na na
#  endif
#endif

/* Wrapper for callback function */

SV *callback_fn;

void callback_wrapper(u_char *user, const struct pcap_pkthdr *h,
	const u_char *pkt)
{
	SV *packet = newSVpv((u_char *)pkt, h->caplen);
	HV *hdr = newHV();
	SV *ref_hdr = newRV_inc((SV*)hdr);

	/* Push arguments onto stack */

        dSP;

	hv_store(hdr, "tv_sec", strlen("tv_sec"), newSViv(h->ts.tv_sec), 0);
	hv_store(hdr, "tv_usec", strlen("tv_usec"), newSViv(h->ts.tv_usec), 0);
	hv_store(hdr, "caplen", strlen("caplen"), newSViv(h->caplen), 0);
	hv_store(hdr, "len", strlen("len"), newSViv(h->len), 0);

        PUSHMARK(sp);
        XPUSHs((SV*)user);
	XPUSHs(ref_hdr);
        XPUSHs(packet);
        PUTBACK;

	/* Call perl function */

        perl_call_sv (callback_fn, G_DISCARD);

	/* Decrement refcount to temp SVs */

        SvREFCNT_dec(packet);
	SvREFCNT_dec(hdr);
	SvREFCNT_dec(ref_hdr);
}

MODULE = Net::Pcap	PACKAGE = Net::Pcap	PREFIX = pcap_

PROTOTYPES: DISABLE

void
pcap_findalldevs(err, ... )
  SV * err

  PREINIT:
    char *errbuf = safemalloc(PCAP_ERRBUF_SIZE);
    SV *err_sv = SvRV(err);
  PPCODE:  
    if (items>2)
      croak("Usage: Net::Pcap::findalldevs(err, [descr])");
    if ( SvROK(err) && ((items==1) || (SvROK(ST(1)) && (SvTYPE(SvRV(ST(1))) == SVt_PVHV))) ) {
      pcap_if_t *alldevs, *d;
      HV *hv;
      if (items==2)
        hv = (HV *)SvRV(ST(1));
      if (pcap_findalldevs(&alldevs, errbuf)==0) {
        for(d=alldevs; d; d=d->next){
          XPUSHs(sv_2mortal(newSVpv(d->name, 0)));
          if (items==2) {          
            if (d->description)
              hv_store(hv, d->name, strlen(d->name), newSVpv(d->description, 0),0);
            else
              hv_store(hv, d->name, strlen(d->name), newSVpv("No description available", 0),0);
          }             
        }
        pcap_freealldevs(alldevs);
        err_sv = &PL_sv_undef;
      }
      else {
        sv_setpv(err_sv, errbuf);
      }
    }
    else {
      if (!SvROK(err)) croak("arg1 not a reference");
      if (items==2 && (!SvROK(ST(1)) || (SvTYPE(SvRV(ST(1))) != SVt_PVHV))) croak("arg2 not a hash reference");			
    }
    safefree(errbuf);
  

#ifdef _WIN32

char *
pcap_lookupdev(err)
	SV *err

	CODE:
	  char *errbuf = safemalloc(PCAP_ERRBUF_SIZE);
		SV *err_sv = SvRV(err);
		RETVAL = pcap_lookupdev(errbuf);
		if (SvROK(err)) {
		  int length = lstrlenW((PWSTR)RETVAL)+2;
		  char *r = safemalloc(length);  // Conversion Unicode to ANSI
		  WideCharToMultiByte(CP_ACP, 0, (PWSTR)RETVAL, -1, r, length, NULL, NULL);	
		  lstrcpyA(RETVAL, r);
		  safefree(r);
		  if (RETVAL == NULL) {
			  sv_setpv(err_sv, errbuf);
		  } 
		  else {
			  err_sv = &PL_sv_undef;
		  }
			safefree(errbuf);
		}
		else
			croak("arg1 not a hash ref");

	OUTPUT:
		RETVAL
		err

#else

char *
pcap_lookupdev(err)
	SV *err

	CODE:
		if (SvROK(err)) {
			char *errbuf = safemalloc(PCAP_ERRBUF_SIZE);
			SV *err_sv = SvRV(err);

			RETVAL = pcap_lookupdev(errbuf);

			if (RETVAL == NULL) {
				sv_setpv(err_sv, errbuf);
			} else {
				err_sv = &PL_sv_undef;
			}

			safefree(errbuf);
		} else
			croak("arg1 not a hash ref");

	OUTPUT:
		RETVAL
		err

#endif

int
pcap_lookupnet(device, net, mask, err)
	char *device
	SV *net
	SV *mask
	SV *err

	CODE:
		if (SvROK(net) && SvROK(mask) && SvROK(err)) {
			char *errbuf = safemalloc(PCAP_ERRBUF_SIZE);
			unsigned int netp, maskp;
			SV *net_sv = SvRV(net);
			SV *mask_sv = SvRV(mask);
			SV *err_sv = SvRV(err);

			RETVAL = pcap_lookupnet(device, &netp, &maskp, errbuf);

			netp = ntohl(netp);
			maskp = ntohl(maskp);

			if (RETVAL != -1) {
				sv_setiv(net_sv, netp);
				sv_setiv(mask_sv, maskp);
				err_sv = &PL_sv_undef;
			} else {
				sv_setpv(err_sv, errbuf);
			}

			safefree(errbuf);

		} else {
			if (!SvROK(net)) croak("arg2 not a reference");
			if (!SvROK(mask)) croak("arg3 not a reference");
			if (!SvROK(err)) croak("arg4 not a reference");
		}

	OUTPUT:
		net
		mask
		err
		RETVAL

pcap_t *
pcap_open_live(device, snaplen, promisc, to_ms, err)
	char *device
	int snaplen
	int promisc
	int to_ms
	SV *err;

	CODE:
		if (SvROK(err)) {
			char *errbuf = safemalloc(PCAP_ERRBUF_SIZE);
			SV *err_sv = SvRV(err);

			RETVAL = pcap_open_live(device, snaplen, promisc,
						to_ms, errbuf);

			if (RETVAL == NULL) {
				sv_setpv(err_sv, errbuf);
			} else {
				err_sv = &PL_sv_undef;
			}

			safefree(errbuf);
		} else
			croak("arg5 not a reference");

	OUTPUT:
		err
		RETVAL


pcap_t *
pcap_open_offline(fname, err)
	char *fname
	SV *err

	CODE:
		if (SvROK(err)) {
			char *errbuf = safemalloc(PCAP_ERRBUF_SIZE);
			SV *err_sv = SvRV(err);

			RETVAL = pcap_open_offline(fname, errbuf);

			if (RETVAL == NULL) {
				sv_setpv(err_sv, errbuf);
			} else {
				err_sv = &PL_sv_undef;
			}

			safefree(errbuf);
		} else
			croak("arg2 not a reference");

	OUTPUT:
		err
		RETVAL

pcap_dumper_t *
pcap_dump_open(p, fname)
	pcap_t *p
	char *fname

int
pcap_loop(p, cnt, callback, user)
	pcap_t *p
	int cnt
	SV *callback
	SV *user

	CODE:
		callback_fn = newSVsv(callback);
		user = newSVsv(user);

		RETVAL = pcap_loop(p, cnt, callback_wrapper, (u_char *)user);

		SvREFCNT_dec(user);
		SvREFCNT_dec(callback_fn);

	OUTPUT:
		RETVAL

int
pcap_dispatch(p, cnt, callback, user)
	pcap_t *p
	int cnt
	SV *callback
	SV *user

	CODE:
		callback_fn = newSVsv(callback);
		user = newSVsv(user);

		RETVAL = pcap_dispatch(p, cnt, callback_wrapper,
				       (u_char *)user);

		SvREFCNT_dec(user);
		SvREFCNT_dec(callback_fn);

	OUTPUT:
		RETVAL

void
pcap_close(p)
	pcap_t *p

void
pcap_dump_close(p)
	pcap_dumper_t *p

int
pcap_datalink(p)
	pcap_t *p

int
pcap_snapshot(p)
	pcap_t *p

int
pcap_is_swapped(p)
	pcap_t *p

int
pcap_major_version(p)
	pcap_t *p

int
pcap_minor_version(p)
	pcap_t *p

char *
pcap_geterr(p)
	pcap_t *p

char *
pcap_strerror(error)
	int error

int
pcap_setnonblock(p, nonblock, err)
    pcap_t * p
    int    nonblock
    SV *   err
  CODE:
    if (SvROK(err)) {
			char *errbuf = safemalloc(PCAP_ERRBUF_SIZE);
			SV *err_sv = SvRV(err);
      RETVAL = pcap_setnonblock(p, nonblock, errbuf);
      if (RETVAL != -1) {
				err_sv = &PL_sv_undef;
			} else {
				sv_setpv(err_sv, errbuf);
			}
			safefree(errbuf);
		} else croak("arg3 not a reference");
OUTPUT:
  err
  RETVAL

int
pcap_getnonblock(p, err)
    pcap_t *    p
    SV *   err
  CODE:
    if (SvROK(err)) {
			char *errbuf = safemalloc(PCAP_ERRBUF_SIZE);
			SV *err_sv = SvRV(err);
      RETVAL = pcap_getnonblock(p, errbuf);
      if (RETVAL != -1) {
				err_sv = &PL_sv_undef;
			} else {
				sv_setpv(err_sv, errbuf);
			}
			safefree(errbuf);
		} else croak("arg2 not a reference");
OUTPUT:
  err
  RETVAL

int
pcap_compile(p, fp, str, optimize, mask)
	pcap_t *p
	SV *fp;
	char *str
	int optimize
	bpf_u_int32 mask

	CODE:
		if (SvROK(fp)) {
			// struct_bpf_program *real_fp = safemalloc(sizeof(fp)); Bug here
			struct_bpf_program *real_fp = safemalloc(sizeof(struct_bpf_program));

			RETVAL = pcap_compile(p, real_fp, str, optimize, mask);

			sv_setref_pv(SvRV(ST(1)), "struct_bpf_programPtr", (void *)real_fp);
		} else
			croak("arg2 not a reference");

	OUTPUT:
		fp
		RETVAL

int
pcap_setfilter(p, fp)
	pcap_t *p
	struct_bpf_program *fp

	
MODULE = Net::Pcap PACKAGE = struct_bpf_programPtr

void
DESTROY(fp)
    struct_bpf_program *fp
  CODE:
    pcap_freecode(fp);
	
MODULE = Net::Pcap	PACKAGE = Net::Pcap	PREFIX = pcap_

int
pcap_fileno(p)
	pcap_t *p

void
pcap_perror(p, prefix)
	pcap_t *p
	char *prefix

int
pcap_stats(p, ps)
	pcap_t *p;
	SV *ps;

	CODE:
		/* Call pcap_stats() function */

		if (SvROK(ps) && (SvTYPE(SvRV(ps)) == SVt_PVHV)) {
			struct pcap_stat real_ps;
			HV *hv;

			RETVAL = pcap_stats(p, &real_ps);

			/* Copy pcap_stats fields into hash */

			hv = (HV *)SvRV(ps);

			hv_store(hv, "ps_recv", strlen("ps_recv"),
		                 newSViv(real_ps.ps_recv), 0);
			hv_store(hv, "ps_drop", strlen("ps_drop"),
			         newSViv(real_ps.ps_drop), 0);
			hv_store(hv, "ps_ifdrop", strlen("ps_ifdrop"),
		                 newSViv(real_ps.ps_ifdrop), 0);

		} else croak("arg 2 not a hash ref");

	OUTPUT:
		RETVAL

FILE *
pcap_file(p)
	pcap_t *p

void
pcap_dump(p, h, sp)
	pcap_dumper_t *p
	SV *h
	SV *sp

	CODE:
		/* Check h (packet header) is a hashref */

		if (SvROK(h) && (SvTYPE(SvRV(h)) == SVt_PVHV)) {
		        struct pcap_pkthdr real_h;
			char *real_sp;
			HV *hv;
			SV **sv;

			memset(&real_h, '\0', sizeof(real_h));

			/* Copy from hash to pcap_pkthdr */

			hv = (HV *)SvRV(h);

			sv = hv_fetch(hv, "tv_sec", strlen("tv_sec"), 0);
			if (sv != NULL) {
				real_h.ts.tv_sec = SvIV(*sv);
			}

			sv = hv_fetch(hv, "tv_usec", strlen("tv_usec"), 0);
			if (sv != NULL) {
				real_h.ts.tv_usec = SvIV(*sv);
			}

			sv = hv_fetch(hv, "caplen", strlen("caplen"), 0);
			if (sv != NULL) {
			        real_h.caplen = SvIV(*sv);
		        }

			sv = hv_fetch(hv, "len", strlen("len"), 0);
			if (sv != NULL) {
			        real_h.len = SvIV(*sv);
			}

			real_sp = SvPV(sp, PL_na);

			/* Call pcap_dump() */

			pcap_dump((u_char *)p, &real_h, real_sp);

		} else croak("arg2 not a hash ref");

SV *
pcap_next(p, h)
        pcap_t *p
	SV *h

	CODE:
		if (SvROK(h) && (SvTYPE(SvRV(h)) == SVt_PVHV)) {
			struct pcap_pkthdr real_h;
			const u_char *result;
			HV *hv;

			memset(&real_h, '\0', sizeof(real_h));

			result = pcap_next(p, &real_h);

			hv = (HV *)SvRV(h);

			if (result != NULL) {

				hv_store(hv, "tv_sec", strlen("tv_sec"),
					 newSViv(real_h.ts.tv_sec), 0);
				hv_store(hv, "tv_usec", strlen("tv_usec"),
					 newSViv(real_h.ts.tv_usec), 0);
				hv_store(hv, "caplen", strlen("caplen"),
					 newSViv(real_h.caplen), 0);
				hv_store(hv, "len", strlen("len"),
					 newSViv(real_h.len), 0);

				RETVAL = newSVpv((char *)result,
						 real_h.caplen);
			} else
				RETVAL = &PL_sv_undef;

		} else croak("arg2 not a hash ref");

	OUTPUT:
	        h
		RETVAL


#ifdef _WINPCAP                 /* ****** WinPcap specific functions */

int 
pcap_createsrcstr(source, type, host, port, name, err)
    SV *    source 
    int     type  
    char *  host 
    char *  port  
    char *  name
    SV *    err
  CODE:
    if (SvROK(source) && SvROK(err)) {
			char *errbuf = safemalloc(PCAP_ERRBUF_SIZE);
			char *sourcebuf = safemalloc(PCAP_BUF_SIZE);
			SV *err_sv = SvRV(err);
			SV *source_sv = SvRV(source);
			
			RETVAL = pcap_createsrcstr(sourcebuf, type, host, port, name, errbuf);
			
			if (RETVAL != -1) {
				sv_setpv(source_sv, sourcebuf);
				err_sv = &PL_sv_undef;
			} else {
				sv_setpv(err_sv, errbuf);
			}

			safefree(errbuf);
			safefree(sourcebuf);

		} else {
			if (!SvROK(source)) croak("arg1 not a reference");
			if (!SvROK(err)) croak("arg6 not a reference");
		}

	OUTPUT:
		source
		err
		RETVAL


int 
pcap_parsesrcstr(source, type, host, port, name, err); 
    char * source  
    SV *   type 
    SV *   host 
    SV *   port  
    SV *   name 
    SV *   err 
  CODE:
    if ( !SvROK(type) ) croak("arg2 not a reference");   
    if ( !SvROK(host) ) croak("arg3 not a reference");  
    if ( !SvROK(port) ) croak("arg4 not a reference");
    if ( !SvROK(name) ) croak("arg5 not a reference");
    if ( !SvROK(err) )  croak("arg6 not a reference");
    else {  
      int rtype;
      char *hostbuf = safemalloc(PCAP_BUF_SIZE);
      char *portbuf = safemalloc(PCAP_BUF_SIZE);
      char *namebuf = safemalloc(PCAP_BUF_SIZE);
      char *errbuf  = safemalloc(PCAP_ERRBUF_SIZE);
      SV *type_sv = SvRV(type);
      SV *host_sv = SvRV(host);
      SV *port_sv = SvRV(port);
      SV *name_sv = SvRV(name);    
      SV *err_sv = SvRV(err);    

      RETVAL = pcap_parsesrcstr(source, &rtype, hostbuf, portbuf, namebuf, errbuf);
         
      if (RETVAL != -1) {
          sv_setiv(type_sv, rtype);
  				sv_setpv(host_sv, hostbuf);
  				sv_setpv(port_sv, portbuf);
  				sv_setpv(name_sv, namebuf);				
  				err_sv = &PL_sv_undef;
  			} else {
  				sv_setpv(err_sv, errbuf);
  			}
      safefree(hostbuf);
      safefree(portbuf);
      safefree(namebuf);
  		safefree(errbuf);
		}
		OUTPUT:
		type
		host
		port
		name
		err
		RETVAL
		
		
pcap_t *
pcap_open(source, snaplen, flags, read_timeout, auth, err)
    char * source
    int snaplen
    int flags
    int read_timeout
    SV * auth
    SV * err
  CODE:      
    if (!SvROK(err)) croak("arg6 not a reference");   
    if ( !SvOK(auth) || (SvOK(auth) && SvROK(auth) && (SvTYPE(SvRV(auth)) == SVt_PVHV)) ){      
      struct pcap_rmtauth real_auth;
      struct pcap_rmtauth * preal_auth;
      char *errbuf = safemalloc(PCAP_ERRBUF_SIZE);
      SV *err_sv = SvRV(err);
      
      if (!SvOK(auth)) {     /* if auth (struct pcap_rmtauth) is undef */
        preal_auth = NULL;
      }      
      else {                 /* auth (struct pcap_rmtauth) is a hashref */  
  			HV *hv;
  			SV **sv;
  
  			memset(&real_auth, '\0', sizeof(real_auth));
  
  			/* Copy from hash to pcap_rmtauth */
  			hv = (HV *)SvRV(auth);
  
  			sv = hv_fetch(hv, "type", strlen("type"), 0);
  			if (sv != NULL) {
  				real_auth.type = SvIV(*sv);
  			}
  
  			sv = hv_fetch(hv, "username", strlen("username"), 0);
  			if (sv != NULL) {
  				real_auth.username = SvPV(*sv, PL_na);
  			}
  
  			sv = hv_fetch(hv, "password", strlen("password"), 0);
  			if (sv != NULL) {
  				real_auth.password = SvPV(*sv, PL_na);
  			}
  			preal_auth = &real_auth;
  		}
  		
  	  RETVAL = pcap_open(source, snaplen, flags, read_timeout, preal_auth, errbuf); 
  	   	   	  
  	  if (RETVAL == NULL) {
  	    sv_setpv(err_sv, errbuf);				
			} else {
			  err_sv = &PL_sv_undef;
			}  	  
  	  safefree(errbuf);
		}
		else croak("arg5 not a reference to a hash");
				
  OUTPUT:
    RETVAL
    err
    
    

#endif


#ifdef _WIN32                   /* ****** Win32 specific functions */

int
pcap_setbuff (p, dim)
    pcap_t *p
    int dim
  CODE:
    RETVAL = pcap_setbuff(p, dim);
  OUTPUT:
    RETVAL


int
pcap_setmode (p, mode)
    pcap_t *p
    int mode
  CODE:
    RETVAL = pcap_setmode(p, mode);
  OUTPUT:
    RETVAL
    
    
int
pcap_setmintocopy(p, size) 
    pcap_t *p
    int size
  CODE:
    RETVAL = pcap_setmintocopy(p, size);
  OUTPUT:
    RETVAL
    

void
pcap_getevent(p)
    pcap_t *p
  PREINIT:
    unsigned int h;  
  PPCODE:
    h = (unsigned int) pcap_getevent(p);  
    ST(0) = sv_newmortal();
	  sv_setref_iv(ST(0), "Win32::Event", h);
    XSRETURN(1);

 
int 
pcap_sendpacket(p, buf)
    pcap_t * p
    SV *  buf
  CODE:
    RETVAL = pcap_sendpacket(p, SvPVX(buf), sv_len(buf));  
  OUTPUT:
    RETVAL


pcap_send_queue * 
pcap_sendqueue_alloc(memsize)
    u_int memsize
  
MODULE = Net::Pcap PACKAGE = pcap_send_queuePtr

void
DESTROY(queue)
    pcap_send_queue * queue
  CODE:
    pcap_sendqueue_destroy(queue);
    	
MODULE = Net::Pcap	PACKAGE = Net::Pcap	PREFIX = pcap_
        
int
pcap_sendqueue_queue(queue, h, p)
    pcap_send_queue * queue
    SV *h
	  SV *p
	CODE:
		/* Check h (packet header) is a hashref */

		if (SvROK(h) && (SvTYPE(SvRV(h)) == SVt_PVHV)) {
		  struct pcap_pkthdr real_h;
			char *real_p;
			HV *hv;
			SV **sv;

			memset(&real_h, '\0', sizeof(real_h));

			/* Copy from hash to pcap_pkthdr */

			hv = (HV *)SvRV(h);

			sv = hv_fetch(hv, "tv_sec", strlen("tv_sec"), 0);
			if (sv != NULL) {
				real_h.ts.tv_sec = SvIV(*sv);
			}

			sv = hv_fetch(hv, "tv_usec", strlen("tv_usec"), 0);
			if (sv != NULL) {
				real_h.ts.tv_usec = SvIV(*sv);
			}

			sv = hv_fetch(hv, "caplen", strlen("caplen"), 0);
			if (sv != NULL) {
			  real_h.caplen = SvIV(*sv);
		  }

			sv = hv_fetch(hv, "len", strlen("len"), 0);
			if (sv != NULL) {
			  real_h.len = SvIV(*sv);
			}

			real_p = SvPV(p, PL_na);

			/* Call pcap_sendqueue_queue() */
      
      RETVAL = pcap_sendqueue_queue(queue, &real_h, (unsigned char *) real_p);

		} else croak("arg2 not a hash ref");
	OUTPUT:
	  RETVAL	
  
u_int
pcap_sendqueue_transmit(p, queue, sync)
    pcap_t *p
    pcap_send_queue * queue
    int sync

#endif
